#ifndef RUMD_DATA_H
#define RUMD_DATA_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_io_helpers.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <cstring>
#include <cassert>
#include <cerrno>
#include <cmath>
#include <zlib.h>


class Particle {
 public:
  unsigned int MyType;
  float  x,  y,  z;
  float vx, vy, vz;
  float fx, fy, fz;
  int Imx, Imy, Imz;
  float Ui, Wi;

  double Rsq(Particle *P, float Lx, float Ly, float Lz) {

    double dx = x - P->x + Lx*(Imx - P->Imx);
    double dy = y - P->y + Ly*(Imy - P->Imy);
    double dz = z - P->z + Lz*(Imz - P->Imz);

    return dx*dx + dy*dy + dz*dz;
  }

  double RsqMinImage(Particle *P, float Lx, float Ly, float Lz, double invLx,  double invLy,  double invLz) {

    double dx = x - P->x;
    double dy = y - P->y;
    double dz = z - P->z;

    dx -= Lx*rintf(dx*invLx);
    dy -= Ly*rintf(dy*invLy);
    dz -= Lz*rintf(dz*invLz);

    return dx*dx + dy*dy + dz*dz;
  }

  double Fsq(Particle *P, float Lx, float Ly, float Lz, float q) {
    double Sum = 0.0;

    Sum += cos( (x - P->x + Lx*(Imx - P->Imx)) * q );
    Sum += cos( (y - P->y + Ly*(Imy - P->Imy)) * q );
    Sum += cos( (z - P->z + Lz*(Imz - P->Imz)) * q );

    return Sum/3.0f;
  }

};

class Conf {
 private:
  // protect from default copy construction and copy assignment 
  Conf(const Conf&);
  Conf& operator=(const Conf&);
 public:
  unsigned int num_part;
  Particle *P;
  ConfMetaData  metaData;

  Conf(): 
    num_part(0), P(0), metaData(ConfMetaData())
  {
  }

  void SetNumberOfParticles(unsigned int num_part_set) {

    assert(num_part_set > 0);
    if (num_part==0) {

      // Allocate particles if not done already
      num_part = num_part_set;
      P = new Particle[num_part];
    } else {

      // If already allocated, check that number didn't change
      assert(num_part_set==num_part);
    }
  }

  ~Conf() {
    if(P)
      delete[] P;
  }

  void Read(gzFile gp, char *filename, int verbose) {

    char strWorkspace[256];
    unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
    assert(maxLen == 256);
    int nItems;
    unsigned int set_num_part=0;
    char *ok; // for checking success of fgets

    // read the first line
    ok = gzgets(gp, strWorkspace, maxLen);
    if(!ok) {
      fprintf(stderr, "ReadConf: Error reading first line of file %s\n", filename);
      exit(1);
    }

    nItems = sscanf(strWorkspace, "%u", &set_num_part);
    if(nItems != 1) {
      fprintf (stderr, "Read: Error reading number of particles from file %s\n",filename);
      fprintf (stderr, "Read: %s\n",strWorkspace);
      exit(1);
    }

    // Set everything associated with number of particles, incl memory alloc
    assert(set_num_part > 0);
    SetNumberOfParticles(set_num_part);

    // read the comment line
    ok = gzgets(gp, strWorkspace, maxLen);
    if(!ok || strlen(strWorkspace) == maxLen - 1) {
      fprintf(stderr, "Read: Error reading comment line in file %s\n",filename);
      exit(1);
    }
    else {
      // replace final newline with a zero
      int lenCmtLine = strlen(strWorkspace);
      assert(strcmp(&(strWorkspace[lenCmtLine-1]),"\n") == 0);
      strWorkspace[lenCmtLine-1] = 0;
    }

    // read data from comment line into metaData
    ConfMetaReader(strWorkspace, &metaData, verbose);

    // read the per-particle data
    for ( unsigned int i = 0; i < num_part; i++ ){

      ok = gzgets(gp, strWorkspace, maxLen);
      if(!ok || strlen(strWorkspace) == maxLen - 1) {
	fprintf(stderr, "Error reading line %u from configuration file\n", i);
	exit(1);
      }
      int offset = 0;
      int partial_offset = 0;

      // type and positions
      sscanf(strWorkspace+offset, "%u %f %f %f%n", &P[i].MyType, &P[i].x, &P[i].y, &P[i].z, &partial_offset);
      offset += partial_offset;
      //printf("%d %f %f %f", P[i].MyType, P[i].x, P[i].y, P[i].z);
      assert(P[i].MyType < metaData.numTypes);

      if(metaData.has_images) {
	sscanf(strWorkspace+offset, "%d %d %d%n", &P[i].Imx, &P[i].Imy, &P[i].Imz, &partial_offset);
        offset += partial_offset;
        //printf(" %d %d %d\n", P[i].Imx, P[i].Imy, P[i].Imz);
      } else {
	P[i].Imx=0; P[i].Imy=0; P[i].Imz=0;
      }

      if(metaData.has_velocities) {
        sscanf(strWorkspace+offset, "%f %f %f%n", &P[i].vx, &P[i].vy, &P[i].vz, &partial_offset);
        offset += partial_offset;
      }

      if(metaData.has_forces) {
	sscanf(strWorkspace+offset, "%f %f %f%n", &P[i].fx, &P[i].fy, &P[i].fz, &partial_offset);
        offset += partial_offset;
      }

      if(metaData.has_pot_energies) {
	sscanf(strWorkspace+offset, "%f%n", &P[i].Ui, &partial_offset);
        offset += partial_offset;
      }

      if(metaData.has_virials) {
	sscanf(strWorkspace+offset, "%f%n", &P[i].Wi, &partial_offset);
        offset += partial_offset;
      }

      if(metaData.has_moleculeIndices) {
	int dummy; // Trond will implement an array for this
	sscanf(strWorkspace+offset, "%d%n", &dummy, &partial_offset);
        offset += partial_offset;
      }
    }
    //ok = fgets(strWorkspace, maxLen, fc); // apparently necessary...
  }

  void ReadOld(FILE *fc, char *filename, int verbose) {

    char strWorkspace[256];
    unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
    assert(maxLen == 256);
    int nItems;
    unsigned int set_num_part;
    char *ok; // for checking success of fgets

    // read the first line
    ok = fgets(strWorkspace, maxLen, fc);
    if(!ok) fprintf(stderr, "Read: Error reading first line of file %s\n", filename);
    // read number of particles--for now just compare to existing

    nItems = sscanf(strWorkspace, "%u", &set_num_part);
    if(nItems != 1) {
      fprintf (stderr, "Read: Error reading number of particles from file %s\n",filename);
      fprintf (stderr, "Read: %s\n",strWorkspace);
      exit(1);
    }

    // Set everything associated with number of particles, incl memory alloc
    SetNumberOfParticles(set_num_part);

    // read the comment line
    ok = fgets(strWorkspace, maxLen, fc);
    if(!ok || strlen(strWorkspace) == maxLen - 1) {
      fprintf(stderr, "Read: Error reading comment line in file %s\n",filename);
      exit(1);
    }
    else {
      // replace final newline with a zero
      int lenCmtLine = strlen(strWorkspace);
      assert(strcmp(&(strWorkspace[lenCmtLine-1]),"\n") == 0);
      strWorkspace[lenCmtLine-1] = 0;
    }

    // read data from comment line into metaData
    ConfMetaReader(strWorkspace, &metaData, verbose);

    // read the per-particle data
    for ( unsigned int i = 0; i < num_part; i++ ){

      // type and positions
      fscanf(fc, "%u %f %f %f", &P[i].MyType, &P[i].x, &P[i].y, &P[i].z);
      //printf("%d %f %f %f", P[i].MyType, P[i].x, P[i].y, P[i].z);
      assert(P[i].MyType < metaData.numTypes);

      if(metaData.has_images) {
	fscanf(fc, "%d %d %d", &P[i].Imx, &P[i].Imy, &P[i].Imz);
      } else {
	P[i].Imx=0; P[i].Imy=0; P[i].Imz=0;
      }

      if(metaData.has_velocities)
        fscanf(fc, "%f %f %f", &P[i].vx, &P[i].vy, &P[i].vz);

      if(metaData.has_forces)
	fscanf(fc, "%f %f %f", &P[i].fx, &P[i].fy, &P[i].fz);

      if(metaData.has_pot_energies)
	fscanf(fc, "%f", &P[i].Ui);

      if(metaData.has_virials)
	fscanf(fc, "%f", &P[i].Wi);

      if(metaData.has_moleculeIndices) {
	int dummy; // Trond will implement an array for this
	fscanf(fc, "%d", &dummy);
      }
    }
    ok = fgets(strWorkspace, maxLen, fc); // apparently necessary...

  }


  void UiWi(Conf *Ct, double *MeanUi,  double *MeanWi, double *Ui, double *Wi, unsigned int num_types) {

    for (unsigned int i=0; i<num_part; i++) {
      unsigned int Type = P[i].MyType;
      assert(Type<num_types);
      assert(Type==Ct->P[i].MyType);
      Ui[Type] += (P[i].Ui-MeanUi[Type])*(Ct->P[i].Ui-MeanUi[Type]);
      Wi[Type] += (P[i].Wi-MeanWi[Type])*(Ct->P[i].Wi-MeanWi[Type]);
    }
  }

};
#endif
