/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <string>
#include <getopt.h>

#include "md_sample.h"
#include "md_timer.h"
#include "md_device.h"
#include "md_IntegratorNoseHoover.h"

int main( int argc, char* argv[]) {
  int num_part = 1024;
  int parts_per_block = 16;
  int threads_per_part = 1;
  int NVT = 0;                         // 0=NVE 1=NVT (Nose-Hoover)
  int num_io_blocks = 1;
  int io_block_size = 1;
  int write_trajectory = 1;
  unsigned int quenchInterval = 50;
  unsigned int lastQuench = 500;
  int mode = 1; // 0 = generate configuration, 1 = run trajectory and write final info; 2 = run trajectory and and compare final info to previous runs.
  
  float L=1.0;
  float Ti = 1.f;                      // Initial temperature
  float Tf = 1.f;                      // Final temperature
  float dt = 0.001f;
  float skin = 0.3f;
  float Rscal = 1.f;                   // Scaling of initial configuration
  float tau = 0.3f;
  float4 sumU, sumK, sumW, sumS, sumC;
  float stress[6];
  float totalDiffSq = 0.; // for mode 2
  
  bool Tf_assigned = false;
  
  while (1) {
    static struct option long_options[] = {
      {"n", 1, 0, 'n'},
      {"pb", 1, 0, 'p'},
      {"tp", 1, 0, 't'},
      {"thermostat", 1, 0, 'e'},
      {"numIOBlocks", 1, 0, 'o'},
      {"IOBlockSize", 1, 0, 'b'},
      {"writeTraj", 1, 0, 'w'},
      {"qi", 1, 0, 'q'},
      {"mode", 1, 0, 'm'},
      {"L", 1, 0, 'l'},
      {"Ti", 1, 0, 'i'},
      {"Tf", 1, 0, 'f'},
      {"dt", 1, 0, 'd'},
      {"tau", 1, 0, 'u'},
      {"skin", 1, 0, 's'},
      {"Rscal", 1, 0, 'r'},
      {0, 0, 0, 0}
    };
    int option_index = 0;
    int c = getopt_long(argc, argv, "",long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
    case 0:
      printf("option %s", long_options[option_index].name);
      if (optarg)
        printf(" with arg %s", optarg);
        printf("\n");
      break;
    case 'n':
      num_part = atoi(optarg);
      break;
    case 'p':
      parts_per_block = atoi(optarg);
      break;
    case 't':
      threads_per_part = atoi(optarg);
      break;
    case 'e':
      NVT = atoi(optarg);
      break;
    case 'o':
      num_io_blocks = atoi(optarg);
      break;
    case 'b':
      io_block_size = atoi(optarg);
      break; 
    case 'w':
      write_trajectory = atoi(optarg);
      break;
    case 'q':
      quenchInterval = (unsigned int)atoi(optarg);
      break;
    case 'm':
      mode = atoi(optarg);
      break;
    case 'l':
      L = strtof(optarg, 0);
      break;
    case 'i':
      Ti = strtof(optarg, 0);
      break;
    case 'f':
      Tf = strtof(optarg, 0);
      Tf_assigned = true;
      break;
    case 'd':
      dt = strtof(optarg, 0);
      break;
    case 'u':
      tau = strtof(optarg, 0);
      break;
    case 's':
      skin = strtof(optarg, 0);
      break;
    case 'r':
      Rscal = strtof(optarg, 0);
      break;
    case '?':
      exit(EXIT_FAILURE);
      break;
    default:
      printf("?? getopt returned character code 0%o ??\n", c);
    }
  }

  if (!Tf_assigned) Tf = Ti;

  md_device_init(argc, argv);
  
  // Create a new sample
  Sample *S = new Sample(parts_per_block, threads_per_part);
  
  // Create an integrator.
  IntegratorNoseHoover* itg =  new IntegratorNoseHoover(S);
  
  // Setup the integrator parameters.
  itg->SetTimeStep(dt);
  itg->SetRelaxationTime(tau);
  itg->SetTargetTemperature(Ti);
  S->SetIntegrator(itg);
  
  Pot_LJ_12_6 potential; 
  
  // Setup parameters for Lennard-Jones interactions ( Kob & Andersen)
  //             i  j   Sig   Eps  Rcut (in units of Sigma_ij)
  potential.SetParams(0, 0, 1.00, 1.00, 2.50);
  potential.SetParams(0, 1, 0.80, 1.50, 2.50);
  potential.SetParams(1, 0, 0.80, 1.50, 2.50);
  potential.SetParams(1, 1, 0.88, 0.50, 2.50);    
  S->SetPotential(&potential,skin);
  
  S->DumpPotentials();
  S->writeTrajectoryConfs = write_trajectory;;
  
  if(mode == 0){
    printf("Creating random conf\n");
    S->CreateRandomConf(num_part, L, 0.8, 20*Ti);
    S->WriteConf("new_conf.xyz"); // different name from start, must be renamed (to avoid accidental overwriting)
  }
  else{
    S->ReadConf("start.xyz.gz");
  }
  
  itg->AllocateThermostatState();  
  if(!NVT) itg->SetThermostatState(0);
  
  // Calculate forces
  S->UpdateNBlistInit();
  S->CalcF(&potential); 
	
  S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
  float Tt = sumK.y;
  printf("Have initial conf; initial potential energy is %f\n",sumU.w/S->num_part);
  double startU_diff = -1.0, startK_diff = -1.0, startW_diff = -1.0, totalE=-1.0;
  FILE *start_energy_f;
  if(mode == 1) {
    start_energy_f = fopen("start_energies.dat","w");
    if(!start_energy_f){
      fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
      exit(1);
    }
    fprintf(start_energy_f, "%.8f %.8f %.8f %.8f\n",  sumU.w, sumK.w, 
	    sumU.w + sumK.w, sumW.w);
    fclose(start_energy_f);
  }
  else if (mode == 2){
    start_energy_f = fopen("start_energies.dat","r");
    if(!start_energy_f){
      fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
      exit(1);
    }
    
    assert(start_energy_f);
    int f_stat = fscanf(start_energy_f, "%lf %lf %lf %lf", &startU_diff, &startK_diff, &totalE, &startW_diff);
    assert(f_stat == 4);
    startU_diff -= sumU.w;
    startK_diff -= sumK.w;
    startW_diff -= sumW.w;
    startU_diff /= S->num_part;
    startK_diff /= S->num_part;
    startW_diff /= S->num_part;
    totalDiffSq += startU_diff*startU_diff;
    totalDiffSq += startK_diff*startK_diff;
    totalDiffSq += startW_diff*startW_diff;
    fclose(start_energy_f);  
  }
  
  
  // Setup of various stuff
  S->SetIOBlockInfo(io_block_size);
  S->outputFormat.writeInterval = 1;
  /*S->outputFormat.Set("stress_xx", true);
  S->outputFormat.Set("stress_yy", true);
  S->outputFormat.Set("stress_zz", true);
  S->outputFormat.Set("stress_xy", true);*/
  md_timer timer;
  // double eps = 1.e-10;
  
  unsigned long int num_iter = (unsigned long int)num_io_blocks * io_block_size;
  float dT = (Tf - Ti)/float(num_iter);
  
  printf("Doing %lu timesteps:\n", num_iter);
  // execute the kernel(s)
  for( unsigned long int i = 0; i < num_iter; i++) {
    
    if(mode == 0 && !NVT && i%quenchInterval == 0 && i < lastQuench)
      S->ZeroAllVelocities();  
    
    // Calculate forces
    S->UpdateNBlist();
    S->CalcF(&potential);   
    S->NotifyFileWriters(i,itg->GetTimeStep());
    // Integrate velocities and positions
    itg->Integrate();
    
    if(NVT){
      itg->SetTargetTemperature(Ti + float(i)*dT);
      itg->UpdateThermostatState();
    }
  }	
	
  double gpu_time = timer.elapsed();
  printf( "\nProcessing time GPU (s): %f\n", gpu_time);
  printf( " per timestep  GPU (ms): %f\n", 1000.0*gpu_time/num_iter);
  
  // Update forces and velocities
  S->UpdateNBlistInit();
  S->CalcF(&potential);
  S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
  Tt = sumK.y;
  double endU_diff = 0., endK_diff = 0., endW_diff = 0.;
  if (mode == 0) {
    S->WriteConf("start.xyz");
  }
  if(mode == 1) {
    FILE *end_energy_f = fopen("end_energies.dat","w");
    if(!end_energy_f){
      fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
      exit(1);
    }
    
    fprintf(end_energy_f, "%.8f %.8f %.8f %.8f %.8f\n",  sumU.w, sumK.w, sumU.w+sumK.w, sumW.w, gpu_time);
    fprintf(end_energy_f, "nSteps= %lu; dt=%f\n",num_iter, itg->GetTimeStep());
    fclose(end_energy_f);
    S->WriteConf("end.xyz");
  }
  else if(mode == 2) {
    printf("mode 2\n");
    FILE *end_energy_f = fopen("end_energies.dat","r");
    if(!end_energy_f){
      fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
      exit(1);
    }
    
    double totalE, prevGpuTime;
    fscanf(end_energy_f, "%lf %lf %lf %lf %lf", &endU_diff, &endK_diff, &totalE, &endW_diff, &prevGpuTime);
    endU_diff -= sumU.w;
    endK_diff -= sumK.w;
    endW_diff -= sumW.w;
    endU_diff /= S->num_part;
    endK_diff /= S->num_part;
    endW_diff /= S->num_part;
    
    totalDiffSq += endU_diff*endU_diff;
    totalDiffSq += endK_diff*endK_diff;
    totalDiffSq += endW_diff*endW_diff;
    FILE *statusFile = fopen("status.txt","w");
    if(!statusFile){
      fprintf (stderr,"Error opening file status.txt: %s\n", strerror(errno));
      exit(1);
    }
    
	  
    fprintf(statusFile, "Status for N=%d, L=%.2f, pb=%d, tp=%d, ",S->num_part,L,parts_per_block,threads_per_part);
    if(NVT)
      fprintf(statusFile, "NVT : ");
    else
      fprintf(statusFile, "NVE : ");
    if(totalDiffSq < 1.e-6)
      // need to allow for changes in round-off error when different
      // (pb, tp) combinations are used so tolerance should not be too small
      fprintf(statusFile, "OK\n");
    else
      fprintf(statusFile, "FAIL\n");
    
    fprintf(statusFile,"Total difference squared: %g\n",totalDiffSq);
    fprintf(statusFile,"startU_diff: %.8f\n",startU_diff);
    fprintf(statusFile,"startK_diff: %.8f\n",startK_diff);
    fprintf(statusFile,"startW_diff: %.8f\n",startW_diff);
    fprintf(statusFile,"endU_diff: %.8f\n",endU_diff);
    fprintf(statusFile,"endK_diff: %.8f\n",endK_diff);
    fprintf(statusFile,"endW_diff: %.8f\n",endW_diff);
    fprintf(statusFile,"time taken, before/now (s): %f/%f\n",prevGpuTime, gpu_time);
  }
  delete S;
  
}
