/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "md_IntegratorNVU.h"
#include "md_kernel.h"

//////////////////////////////////////////////////////////////////////////
///////////////////////// Constructors ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////

IntegratorNVU::IntegratorNVU(Sample* newSample){
  S = newSample; // Set the sample.
  numberOfBlocks = 1;
  targetPotentialEnergy = 1.f;
  timeStep = 1.f;
  displacementLength = 0.1f;

  AllocateNVUState(numberOfBlocks, targetPotentialEnergy); 
}

IntegratorNVU::~IntegratorNVU(){
  FreeNVUState();
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Class Methods //////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNVU::Integrate(){
  integrateNVUAlgorithm<<< S->grid, S->threads_tp1 >>>( S->num_part, S->particleData.d_r, S->particleData.d_v, S->particleData.d_f, S->particleData.d_im, 
							S->d_L, S->d_invL, S->neighborList.d_ctrl, d_constraintForce, d_displacementLengthCorrection, S->skin );
}

// Constrains the motion on U.
void IntegratorNVU::CalculateConstraintForce(){
  partialSummationConstraintForce<<< S->grid, S->threads_tp1, 2*(S->shared_size_tp1) >>>( S->num_part, S->particleData.d_v, S->particleData.d_f, d_constraintForceNumerator, 
											  d_constraintForceDenominator );

  calculateConstraintForce<<< S->grid, S->threads_tp1 >>>( d_constraintForceNumerator, d_constraintForceDenominator, d_constraintForce, d_totalConstraintForceDenominator );
}

// Step length and potential energy correction. Reuses the summation arrays.
void IntegratorNVU::CalculateNumericalCorrection(){
  partialSummationNumericalCorrection<<< S->grid, S->threads_tp1, 2*(S->shared_size_tp1) >>>( S->num_part, S->particleData.d_v, S->particleData.d_f, d_constraintForce, 
  											      d_totalConstraintForceDenominator, d_constraintForceCorrection, 
											      d_constraintForceNumerator, d_constraintForceDenominator, d_previousPotentialEnergy, 
											      S->num_part * targetPotentialEnergy, displacementLength, 
											      S-> potential->potential_factor );
  
  calculateNumericalCorrection<<< S->grid, S->threads_tp1 >>>( d_constraintForceNumerator, d_constraintForceDenominator, d_displacementLengthCorrection, 
							       d_previousPotentialEnergy, displacementLength);
}

void IntegratorNVU::AllocateNVUState(){
  numberOfBlocks = S->allocatedNumBlocks;
  FreeNVUState();
  AllocateNVUState(numberOfBlocks, targetPotentialEnergy);
}

// Private method.
void IntegratorNVU::AllocateNVUState( int nb, float targetPotentialEnergy ){
  // U(n-1).
  h_previousPotentialEnergy = (float*) malloc( sizeof(float) );
  h_previousPotentialEnergy[0] = S->num_part * targetPotentialEnergy; 
  cudaMalloc( (void**) &d_previousPotentialEnergy, sizeof(float) );  
  cudaMemcpy( d_previousPotentialEnergy, h_previousPotentialEnergy, sizeof(float), cudaMemcpyHostToDevice );
  
  // Constraint force.
  h_constraintForce = (float*) malloc( sizeof(float) );
  cudaMalloc( (void**) &d_constraintForce, sizeof(float) );  
  cudaMalloc( (void**) &d_totalConstraintForceDenominator, sizeof(float) );  
  
  // Numerical corrections CPU.
  h_constraintForceCorrection = (float*) malloc( sizeof(float) );
  h_constraintForceCorrection[0] = 0;
  h_displacementLengthCorrection = (float*) malloc( sizeof(float) );
  h_displacementLengthCorrection[0] = 1.f;
  
  // Numerical corrections GPU.
  cudaMalloc( (void**) &d_constraintForceCorrection, sizeof(float) );  
  cudaMalloc( (void**) &d_displacementLengthCorrection, sizeof(float) );  
  cudaMemcpy( d_displacementLengthCorrection, h_displacementLengthCorrection, sizeof(float), cudaMemcpyHostToDevice );

  // Allocate space on the GPU for the summation of the constraint force.
  cudaMalloc( (void**) &d_constraintForceNumerator, nb * sizeof(float) );  
  cudaMalloc( (void**) &d_constraintForceDenominator, nb * sizeof(float) );  

  // Allocate space for the summation of the system velocity on GPU.
  cudaMalloc( (void**) &d_particleVelocity, nb * sizeof(float4) );  

  // Euclidean configuration.
  h_euclideanConfiguration = (float4*) malloc( S->num_part * sizeof(float4) );
}

void IntegratorNVU::FreeNVUState(){
  free(h_previousPotentialEnergy);
  free(h_constraintForce);
  free(h_constraintForceCorrection);
  free(h_displacementLengthCorrection);
  free(h_euclideanConfiguration);
  cudaFree(d_previousPotentialEnergy);
  cudaFree(d_constraintForce);
  cudaFree(d_totalConstraintForceDenominator);
  cudaFree(d_constraintForceCorrection);
  cudaFree(d_displacementLengthCorrection);
  cudaFree(d_constraintForceNumerator);
  cudaFree(d_constraintForceDenominator);
  cudaFree(d_particleVelocity);
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Set Methods ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNVU::SetTimeStep( float dt ){ timeStep = dt; }
void IntegratorNVU::SetDisplacementLength( float step ){ displacementLength = step; }
void IntegratorNVU::SetTargetPotentialEnergy( float U ){ targetPotentialEnergy = U; };
void IntegratorNVU::SetSystemVelocityToZero(){
  partialSummationVelocity<<< S->grid, S->threads_tp1, S->shared_size >>>( S->num_part, S->particleData.d_v, d_particleVelocity );
  zeroSystemVelocity<<< S->grid, S->threads_tp1, S->shared_size >>>( S->num_part, S->particleData.d_v, d_particleVelocity );
}

void IntegratorNVU::SetEuclideanConfiguration(){
  S->particleData.CopyPosFromDevice();

  for(unsigned int i=0; i < S->num_part; i++){
    h_euclideanConfiguration[i] = S->particleData.h_r[i];
  }
}

void IntegratorNVU::SetEuclideanConfiguration(float4* c){
  for(unsigned int i=0; i < S->num_part; i++){
    h_euclideanConfiguration[i] = c[i];
  }
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Get Methods ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

float IntegratorNVU::GetTimeStep(){ return timeStep; }

float IntegratorNVU::GetSimulationTimeStepSq(){ 
  cudaMemcpy( h_constraintForce, d_constraintForce, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForce[0]; 
}

float IntegratorNVU::GetDisplacementLength(){ return displacementLength; }

float IntegratorNVU::GetSimulationDisplacementLength(){ 
  S->particleData.CopyVelFromDevice();
  
  float dispLength = 0;

  for(unsigned int i=0; i < S->num_part; i++){
    dispLength += S->particleData.h_v[i].x * S->particleData.h_v[i].x + S->particleData.h_v[i].y * S->particleData.h_v[i].y 
      + S->particleData.h_v[i].z * S->particleData.h_v[i].z;
  }

  return sqrtf(dispLength);
} 

float IntegratorNVU::GetConstraintForceCorrection(){ 
  cudaMemcpy( h_constraintForceCorrection, d_constraintForceCorrection, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForceCorrection[0]; 
};

float IntegratorNVU::GetDisplacementLengthCorrection(){ 
  cudaMemcpy( h_displacementLengthCorrection, d_displacementLengthCorrection, sizeof(float), cudaMemcpyDeviceToHost );
  return h_displacementLengthCorrection[0]; 
};

float IntegratorNVU::GetTargetPotentialEnergy(){ return targetPotentialEnergy; };

float IntegratorNVU::GetLagrangeMultiplier(){ 
  cudaMemcpy( h_constraintForce, d_constraintForce, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForce[0] / displacementLength;
} 

float IntegratorNVU::GetEuclideanLength(){ 
  S->particleData.CopyPosFromDevice();
  double distanceSq = 0;

  for(unsigned i=0; i < S->num_part; i++){
    double diffX = h_euclideanConfiguration[i].x - S->particleData.h_r[i].x;
    double diffY = h_euclideanConfiguration[i].y - S->particleData.h_r[i].y;
    double diffZ = h_euclideanConfiguration[i].z - S->particleData.h_r[i].z;

    distanceSq += diffX * diffX + diffY * diffY + diffZ * diffZ;
  }

  return sqrt(distanceSq);
} 

double IntegratorNVU::GetKineticEnergy(){ return -1; } // Not implemented.

//////////////////////////////////////////////////////////////////////////
//////////// Kernel Integration implementation ///////////////////////////
//////////////////////////////////////////////////////////////////////////

/// NVU algorithm: [ ]
__global__ void integrateNVUAlgorithm( unsigned numParticles, float4* position, float4* velocity, float4* force, float4* image, float* boxL, float* invboxL, 
				       int* updateNbList, float* constraintForce, float* displacementLengthCorrection, float nbListSkin ){
  
  if ( MyGP < numParticles ){ 
    float4 my_position = position[MyGP]; float4 my_velocity = velocity[MyGP]; float4 my_force = force[MyGP]; float4 my_image = image[MyGP];
    float L = boxL[0]; float invL = invboxL[0]; 
    
    float cForce = constraintForce[0];
    float dispLengthCorrection = displacementLengthCorrection[0];
    
    // Delta R(i+1/2).
    my_velocity.x = dispLengthCorrection * ( my_velocity.x + cForce * my_force.x );
    my_velocity.y = dispLengthCorrection * ( my_velocity.y + cForce * my_force.y ); 
    my_velocity.z = dispLengthCorrection * ( my_velocity.z + cForce * my_force.z );
    
    // R(i+1).
    my_position.x += my_velocity.x;
    my_position.y += my_velocity.y;
    my_position.z += my_velocity.z;
    
    my_image.x = rintf(my_position.x*invL);
    my_image.y = rintf(my_position.y*invL);
    my_image.z = rintf(my_position.z*invL);
    my_image.w += __powf( my_velocity.x * my_velocity.x + my_velocity.y * my_velocity.y + my_velocity.z * my_velocity.z, 0.5);  
    
    // Periodic boundary. 
    my_position.x -= L*my_image.x;  
    my_position.y -= L*my_image.y;  
    my_position.z -= L*my_image.z;  
    
    // Save the integration in global memory-
    position[MyGP] = my_position; 
    velocity[MyGP] = my_velocity; 
    
    // Avoid calculation in above statements.
    image[MyGP].x += my_image.x;
    image[MyGP].y += my_image.y;
    image[MyGP].z += my_image.z;
    image[MyGP].w = my_image.w;
    
    // Update neighbor list?
    if( my_image.w > 0.5 * nbListSkin ) updateNbList[0] = NumBlocks; 
  }
}

//////////////////////////////////////////////////////////////////////////
///////////////////// Constraint Force ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////

__global__ void calculateConstraintForce( float* constraintForceNumerator,  float* constraintForceDenominator, 
					  float* constraintForceCorrection, float* totalConstraintForceDenominator ){
  if(MyP == 0){
    float constraintFNumerator = constraintForceNumerator[0];
    float constraintFDenominator = constraintForceDenominator[0]; 
    
    // Sum all block contributions.
    for( int i=1; i < NumBlocks; i++ ){
      constraintFNumerator += constraintForceNumerator[i];
      constraintFDenominator += constraintForceDenominator[i];
    }
    
    constraintForceCorrection[0] = ( -2.f * constraintFNumerator ) / constraintFDenominator;
    totalConstraintForceDenominator[0] = constraintFDenominator;
  }
}

// A simpe partial summation.
__global__ void partialSummationConstraintForce( unsigned numParticles, float4* velocity, float4* force, float* constraintForceNumerator, float* constraintForceDenominator ){
  
  extern __shared__ float storageArrayConstraintForce[];
  
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; float4 my_force = force[MyGP]; 
    
    // Items for calculating the constraint force.
    storageArrayConstraintForce[MyP] = my_velocity.x * my_force.x + my_velocity.y * my_force.y + my_velocity.z * my_force.z;
    storageArrayConstraintForce[MyP+PPerBlock] = my_force.x * my_force.x + my_force.y * my_force.y + my_force.z * my_force.z;
  }
  else{
    storageArrayConstraintForce[MyP] = 0;
    storageArrayConstraintForce[MyP+PPerBlock] = 0;
  }    
  
  // Ready for the sum?
  __syncthreads();
  
  // Sum the contributions from this block. Reduces the sum from global memory 1 time.
  if(MyP == 0){
    for( int i=1; i < PPerBlock; i++ ){
      storageArrayConstraintForce[0] += storageArrayConstraintForce[i];
      storageArrayConstraintForce[PPerBlock] += storageArrayConstraintForce[PPerBlock+i];
    }
    
    constraintForceNumerator[MyB] = storageArrayConstraintForce[0];
    constraintForceDenominator[MyB] = storageArrayConstraintForce[PPerBlock];
  }
}

//////////////////////////////////////////////////////////////////////////
////////////////// Normalization Vector //////////////////////////////////
//////////////////////////////////////////////////////////////////////////

__global__ void calculateNumericalCorrection( float* potentialEnergyArray, float* normalizationLengthArray, float* displacementLengthCorrection,
					      float* previousPotentialEnergy, float displacementLength ){
  
  if(MyP == 0){
    float currentPotentialEnergy = potentialEnergyArray[0];
    float normalizationLength = normalizationLengthArray[0]; 
    
    // Sum all block contributions.
    for( int i=1; i < NumBlocks; i++ ){
      currentPotentialEnergy += potentialEnergyArray[i];    
      normalizationLength += normalizationLengthArray[i];
    }
    
    displacementLengthCorrection[0] =  displacementLength / sqrtf(normalizationLength); 
    previousPotentialEnergy[0] = currentPotentialEnergy;
  }
}

__global__ void partialSummationNumericalCorrection( unsigned numParticles, float4* velocity, float4* force, float* constraintForce, float* totalConstraintForceDenominator,  
						     float* constraintForceCorrection, float* potentialEnergyArray, float* normalizationLengthArray, 
						     float* previousPotentialEnergy, float targetPotentialEnergy, float displacementLength, float potentialEnergyFactor ){
  
  extern __shared__ float storageArrayNumericalCorrection[];
  
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; float4 my_force = force[MyGP]; 
    
    float constraintFCorrection = ( previousPotentialEnergy[0] - targetPotentialEnergy ) / totalConstraintForceDenominator[0];
    float constraintF = constraintForce[0] + constraintFCorrection;
    
    constraintForceCorrection[0] = constraintFCorrection;
    constraintForce[0] = constraintF;

    float3 normVector = { my_velocity.x + constraintF * my_force.x, my_velocity.y + constraintF * my_force.y,  my_velocity.z + constraintF * my_force.z };
    
    storageArrayNumericalCorrection[MyP] = normVector.x * normVector.x + normVector.y * normVector.y + normVector.z * normVector.z;
    storageArrayNumericalCorrection[MyP+PPerBlock] = 0.5f * potentialEnergyFactor * my_force.w;
  }
  else{
    storageArrayNumericalCorrection[MyP] = 0;
    storageArrayNumericalCorrection[MyP+PPerBlock] = 0;
  }    
  
  // Ready for the sum?
  __syncthreads();
  
  // Sum the contributions from this block. Reduces the sum from global memory 1 time.
  if(MyP == 0){
    for( int i=1; i < PPerBlock; i++ ){
      storageArrayNumericalCorrection[0] += storageArrayNumericalCorrection[i];
      storageArrayNumericalCorrection[PPerBlock] += storageArrayNumericalCorrection[PPerBlock+i];
    }
    
    normalizationLengthArray[MyB] = storageArrayNumericalCorrection[0];
    potentialEnergyArray[MyB] = storageArrayNumericalCorrection[PPerBlock];
  }
}

//////////////////////////////////////////////////////////////////////////
//////////////// Zero the mean system velocity ///////////////////////////
//////////////////////////////////////////////////////////////////////////

// Zero the system velocity.
__global__ void zeroSystemVelocity( unsigned numParticles, float4* velocity, float4* particleVelocity ){
  
  extern __shared__ float4 storageArraySystemVelocity[];
  
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; 
    
    // Sum the total system velocity.
    if( MyP == 0 ){
      storageArraySystemVelocity[0] = particleVelocity[0];
      
      for( unsigned int i=1; i < NumBlocks; i++ ){
	storageArraySystemVelocity[0].x += particleVelocity[i].x;
	storageArraySystemVelocity[0].y += particleVelocity[i].y;
	storageArraySystemVelocity[0].z += particleVelocity[i].z;
	storageArraySystemVelocity[0].w += particleVelocity[i].w;
      }
    }      
    
    __syncthreads();      
    
    float invTotalMass = 1.f / float(numParticles);
    
    // Subtract
    my_velocity.x -= invTotalMass * storageArraySystemVelocity[0].x;
    my_velocity.y -= invTotalMass * storageArraySystemVelocity[0].y;
    my_velocity.z -= invTotalMass * storageArraySystemVelocity[0].z;
    
    velocity[MyGP] = my_velocity; 
  }
}

__global__ void partialSummationVelocity( unsigned numParticles, float4* velocity, float4* particleVelocity ){
  
  extern __shared__ float4 storageArrayNVU[];
  
  float4 my_velocity = { 0.f, 0.f, 0.f, 0.f };
  
  if( MyGP < numParticles ){
    storageArrayNVU[MyP] = velocity[MyGP];
  }
  else{
    storageArrayNVU[MyP] = my_velocity;
  }
  
  __syncthreads();
  
  if( MyP == 0 ){
    for( unsigned int i=1; i < PPerBlock; i++ ){
      storageArrayNVU[0].x += storageArrayNVU[i].x;
      storageArrayNVU[0].y += storageArrayNVU[i].y;
      storageArrayNVU[0].z += storageArrayNVU[i].z;
      storageArrayNVU[0].w += storageArrayNVU[i].w;
    }
    
    // Store this block contribution.
    particleVelocity[MyB] = storageArrayNVU[0];
  } 
}
