/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "md_IntegratorNoseHoover.h"
#include "md_kernel.h"

//////////////////////////////////////////////////////////////////////////
///////////////////////// Constructors ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////

IntegratorNoseHoover::IntegratorNoseHoover(Sample* newSample){
  S = newSample; // Set the sample.
  numberOfThermostats = 1; // Default number of thermostats allocated.
  targetTemperature = 1.f;
  thermostatMass = 1.f;
  timeStep = 0.0025f;
  thermostatRelaxationTime = 0.2f;

  AllocateThermostatState(numberOfThermostats,0); 
}

IntegratorNoseHoover::~IntegratorNoseHoover(){
  FreeAllocatedMemory();
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Class Methods //////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNoseHoover::Integrate(){
  // Integrate Nose-Hoover dynamics.
  integrateNoseHooverAlgorithm<<< S->grid, S->threads_tp1 >>>( S->num_part, S->particleData.d_r, S->particleData.d_v, S->particleData.d_f, S->particleData.d_im,  
							       S->d_L, S->d_invL, d_thermostatState, S->neighborList.d_ctrl, timeStep, S->skin );
}

void IntegratorNoseHoover::UpdateThermostatState(){
  // Update the thermostat state.
  partialSummationKineticEnergy<<< S->grid, S->threads_tp1, S->shared_size_tp1 >>>( S->num_part, S->particleData.d_v, d_thermostatKineticEnergy ); 
  updateNoseHooverThermostatState<<< S->grid, S->threads_tp1 >>>( d_thermostatKineticEnergy, d_thermostatState, thermostatMass, targetTemperature, S->g, timeStep );
}

void IntegratorNoseHoover::AllocateThermostatState(){
  numberOfThermostats = S->allocatedNumBlocks;
  float state = GetThermostatState();
  FreeAllocatedMemory();
  AllocateThermostatState(numberOfThermostats, state);
}

// Private method.
void IntegratorNoseHoover::AllocateThermostatState( int nb, float initialThermostatState ){
  // Allocate space on the CPU for the thermostat state.
  h_thermostatState = (float*) malloc( nb * sizeof(float) );
  
  // Initialize the state on the CPU.
  for(int i=0; i < nb; i++){ h_thermostatState[0] = initialThermostatState; };
  
  // Allocate space on the GPU for the thermostat state.
  cudaMalloc( (void**) &d_thermostatState, nb * sizeof(float) );
  
  // Initialize the state on the GPU.
  cudaMemcpy( d_thermostatState, h_thermostatState, nb * sizeof(float), cudaMemcpyHostToDevice );
  
  // Allocate space for the summation of kinetic energy on GPU.
  cudaMalloc( (void**) &d_thermostatKineticEnergy, nb * sizeof(float) );  

  // Allocate space for the summation of momentum on GPU.
  cudaMalloc( (void**) &d_particleMomentum, nb * sizeof(float4) );  

  // g is not set before we know N.
  thermostatMass = ( S->g * targetTemperature * thermostatRelaxationTime * thermostatRelaxationTime ) / ( 4.f * M_PI * M_PI ); 
}

// Free all allocated CPU/GPU memory.
void IntegratorNoseHoover::FreeAllocatedMemory(){
  free(h_thermostatState); 
  cudaFree(d_thermostatState); 
  cudaFree(d_thermostatKineticEnergy);
  cudaFree(d_particleMomentum);
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Set Methods ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNoseHoover::SetTimeStep( float dt ){ 
  timeStep = dt; 
}

void IntegratorNoseHoover::SetRelaxationTime( float tau ){ 
  thermostatRelaxationTime = tau; 
  thermostatMass = ( S->g * targetTemperature * thermostatRelaxationTime * thermostatRelaxationTime ) / ( 4.f * M_PI * M_PI ); 
} 

void IntegratorNoseHoover::SetTargetTemperature( float T ){  
  targetTemperature = T;
  thermostatMass = ( S->g * targetTemperature *  thermostatRelaxationTime * thermostatRelaxationTime ) / ( 4.f * M_PI * M_PI ); 
}

void IntegratorNoseHoover::SetThermostatState( float Ps ){
  for(int i=0; i < numberOfThermostats; i++){ h_thermostatState[i] = Ps; };
  cudaMemcpy( d_thermostatState, h_thermostatState, numberOfThermostats * sizeof(float), cudaMemcpyHostToDevice );
}

void IntegratorNoseHoover::SetMomentumToZero(){
  partialSummationMomentum<<< S->grid, S->threads_tp1, S->shared_size >>>( S->num_part, S->particleData.d_v, d_particleMomentum ); 
  zeroTotalMomentum<<< S->grid, S->threads_tp1, S->shared_size >>>( S->num_part, S->particleData.d_v, d_particleMomentum );
}

//////////////////////////////////////////////////////////////////////////
/////////////////////////// Get Methods //////////////////////////////////
//////////////////////////////////////////////////////////////////////////

float IntegratorNoseHoover::GetTimeStep(){ return timeStep; }
float IntegratorNoseHoover::GetRelaxationTime(){ return thermostatRelaxationTime; }
float IntegratorNoseHoover::GetTargetTemperature(){ return targetTemperature; }

float IntegratorNoseHoover::GetThermostatState(){
  cudaMemcpy( h_thermostatState, d_thermostatState, sizeof(float), cudaMemcpyDeviceToHost );
  return h_thermostatState[0];
}

// State: r(t) and v(t-h/2). Return sum( 0.5 m v^2(t) )
// Loops over all particles. May be changed to one particle only.
double IntegratorNoseHoover::GetKineticEnergy(){
  S->particleData.CopyVelFromDevice();
  GetThermostatState();  
  
  double totalKineticEnergy = 0;
  
  for(unsigned int i=0; i < S->num_part; i++){
    double mass = 1.f / S->particleData.h_v[i].w;
    float timeStepTimesinvMass = S->particleData.h_v[i].w * timeStep;
    
    float factor = 0.5f * timeStep * h_thermostatState[0];
    float plus = 1.f / ( 1.f + factor );
    float minus = 1.f - factor;
    
    float velocityFx = plus * ( minus * S->particleData.h_v[i].x + timeStepTimesinvMass * S->particleData.h_f[i].x );
    float velocityFy = plus * ( minus * S->particleData.h_v[i].y + timeStepTimesinvMass * S->particleData.h_f[i].y );
    float velocityFz = plus * ( minus * S->particleData.h_v[i].z + timeStepTimesinvMass * S->particleData.h_f[i].z );
    
    // Velocity: v^2(t-h/2)
    double velocityB = S->particleData.h_v[i].x * S->particleData.h_v[i].x + S->particleData.h_v[i].y * S->particleData.h_v[i].y 
      + S->particleData.h_v[i].z * S->particleData.h_v[i].z;
    
    // Velocity: v^2(t+h/2)
    double velocityF = velocityFx * velocityFx + velocityFy * velocityFy + velocityFz * velocityFz;
    
    // The kinetic energy v^2(t) should be calculated like this in a LFA.
    totalKineticEnergy += mass * ( ( velocityB + velocityF ) / 2.0 );
  }
  
  return 0.5 * totalKineticEnergy;
}

// Return the system total momentum, P.
float4 IntegratorNoseHoover::GetTotalMomentum(){
  S->particleData.CopyVelFromDevice();
  
  float4 totalMomentum = { 0, 0, 0, 0 };
  
  for(unsigned int i=0; i < S->num_part; i++){
    float particleMass = 1.f / S->particleData.h_v[i].w;
    totalMomentum.x += particleMass * S->particleData.h_v[i].x;
    totalMomentum.y += particleMass * S->particleData.h_v[i].y;
    totalMomentum.z += particleMass * S->particleData.h_v[i].z;
    totalMomentum.w += particleMass; // Total mass.
  }

  return totalMomentum;
}

//////////////////////////////////////////////////////////////////////////
//////////// Kernel Integration implementation ///////////////////////////
//////////////////////////////////////////////////////////////////////////

/// NVE algorithm: Leap frog 
/// NVT algorithm: [S. Toxvaerd, Mol. Phys. 72, 159 (1991)]
__global__ void integrateNoseHooverAlgorithm( unsigned numParticles, float4* position, float4* velocity, float4* force, float4* image, 
					      float* boxL, float* invboxL, float* thermostatState, int* updateNbList, float timeStep, float nbListSkin ){
  
  if ( MyGP < numParticles ){ 
    float4 my_position = position[MyGP]; float4 my_velocity = velocity[MyGP]; 
    float4 my_force = force[MyGP]; float4 my_image = image[MyGP];
    float L = boxL[0]; float invL = invboxL[0];
    
    // Thermostat variable. NVE: = 0.
    float factor = 0.5f * thermostatState[0] * timeStep;
    float plus = 1.f / ( 1.f + factor ); // Possibly change to exp(...)
    float minus = 1.f - factor; // Possibly change to exp(...)

    // Update to v(t+h/2).
    my_velocity.x = plus * ( minus * my_velocity.x + my_velocity.w * my_force.x * timeStep ); 
    my_velocity.y = plus * ( minus * my_velocity.y + my_velocity.w * my_force.y * timeStep );
    my_velocity.z = plus * ( minus * my_velocity.z + my_velocity.w * my_force.z * timeStep );
    
    // Update to r(t+h).
    my_position.x += my_velocity.x * timeStep; 
    my_position.y += my_velocity.y * timeStep; 
    my_position.z += my_velocity.z * timeStep; 
    
    // Keep track of the real distance moved.
    my_image.x = rintf(my_position.x*invL);
    my_image.y = rintf(my_position.y*invL);
    my_image.z = rintf(my_position.z*invL);
    // The distance moved between time t and t+h.
    my_image.w += __powf( timeStep * timeStep * ( my_velocity.x * my_velocity.x + my_velocity.y * my_velocity.y + my_velocity.z * my_velocity.z ), 0.5 );    
    
    // Periodic boundary condition.
    my_position.x -= L*my_image.x;
    my_position.y -= L*my_image.y;
    my_position.z -= L*my_image.z;
 
    // Save the integration in global memory.
    position[MyGP] = my_position; 
    velocity[MyGP] = my_velocity;    
    
    // Avoid calculation in above statements.
    image[MyGP].x += my_image.x;
    image[MyGP].y += my_image.y;
    image[MyGP].z += my_image.z;
    image[MyGP].w = my_image.w;
    
    // Update the neighbor list?
    if( my_image.w > 0.5f * nbListSkin ) updateNbList[0] = NumBlocks; 
  }
}

//////////////////////////////////////////////////////////////////////////
//////////////////// Update Nose Hoover State ////////////////////////////
//////////////////////////////////////////////////////////////////////////

// Updates the thermostat state on GPU.
__global__ void updateNoseHooverThermostatState( float* thermostatKineticEnergy, float* thermostatState, float thermostatMass, 
						 float targetTemperature, int degreesOfFreedom, float timeStep ){ 
  // Calculate the total (2 * kinetic energy).
  if ( MyP == 0 ){
    float totalKineticEnergy = 0;       
    
    // Sum all invidividual contributions.
    for ( unsigned int i=0; i < NumBlocks; i++ ){
      totalKineticEnergy += thermostatKineticEnergy[i];
    }
    
    // Update the Nose-Hoover thermostat state.
    thermostatState[MyB] += timeStep * (1.f / thermostatMass) * ( totalKineticEnergy - degreesOfFreedom * targetTemperature );
  }      
}

// A simple summation.
__global__ void partialSummationKineticEnergy( unsigned int numParticles, float4* velocity, float* thermostatKineticEnergy ){
  
  extern __shared__ float storageArrayKineticEnergy[];
  
  // Store 2 * My kinetic energy in shared mem ( my_v.w = 1/mass )
  if ( MyGP < numParticles ){ 
    float4 my_velocity = velocity[MyGP]; 
    
    // m * v^2.
    storageArrayKineticEnergy[MyP] = ( my_velocity.x * my_velocity.x + my_velocity.y * my_velocity.y + my_velocity.z * my_velocity.z ) / my_velocity.w;
  }
  else{
    storageArrayKineticEnergy[MyP] = 0;
  }
  
  __syncthreads();
  
  // One thread in each block does the summation.
  if ( MyP == 0 ){
    for ( unsigned int i=1; i < PPerBlock; i++ ){
      storageArrayKineticEnergy[0] += storageArrayKineticEnergy[i];
    }
    
    thermostatKineticEnergy[MyB] = storageArrayKineticEnergy[0];
  } 
}

//////////////////////////////////////////////////////////////////////////
//////////////////// Set Momentum to Zero ////////////////////////////////
//////////////////////////////////////////////////////////////////////////

// Sets the total momentum of the system to zero.
__global__ void zeroTotalMomentum( unsigned int numParticles, float4* velocity, float4* particleMomentum ){
  
  extern __shared__ float4 storageArrayTotalMomentum[];
  
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; 
    
    // Sum the total momentum.
    if( MyP == 0 ){
      storageArrayTotalMomentum[0] = particleMomentum[0];
      
      for( unsigned int i=1; i < NumBlocks; i++ ){
	storageArrayTotalMomentum[0].x += particleMomentum[i].x;
	storageArrayTotalMomentum[0].y += particleMomentum[i].y;
	storageArrayTotalMomentum[0].z += particleMomentum[i].z;
	storageArrayTotalMomentum[0].w += particleMomentum[i].w;
      }
    }      
    
    __syncthreads();      
    
    // Subtract off total momentum (convert to velocity).
    my_velocity.x -= storageArrayTotalMomentum[0].x / storageArrayTotalMomentum[0].w;
    my_velocity.y -= storageArrayTotalMomentum[0].y / storageArrayTotalMomentum[0].w;
    my_velocity.z -= storageArrayTotalMomentum[0].z / storageArrayTotalMomentum[0].w;
    
    velocity[MyGP] = my_velocity;
  }
}

// A simple summation.
__global__ void partialSummationMomentum( unsigned int numParticles, float4* velocity, float4* particleMomentum ){
  
  extern __shared__ float4 storageArrayMomentum[];
  
  float4 my_momentum = { 0.f, 0.f, 0.f, 0.f };
  
  if( MyGP < numParticles ){
    my_momentum = velocity[MyGP]; 
    
    // Convert to momenta and mass.
    my_momentum.x /=  my_momentum.w;
    my_momentum.y /=  my_momentum.w;
    my_momentum.z /=  my_momentum.w;
    my_momentum.w = 1.f / my_momentum.w;
    
    // Store momenta in shared mem.
    storageArrayMomentum[MyP] = my_momentum;
  }
  else{
    storageArrayMomentum[MyP] = my_momentum;
  }
  
  __syncthreads();
  
  if( MyP == 0 ){
    for( unsigned int i=1; i < PPerBlock; i++ ){
      storageArrayMomentum[0].x += storageArrayMomentum[i].x;
      storageArrayMomentum[0].y += storageArrayMomentum[i].y;
      storageArrayMomentum[0].z += storageArrayMomentum[i].z;
      storageArrayMomentum[0].w += storageArrayMomentum[i].w;
    }
    
    // Store this block contribution.
    particleMomentum[MyB] = storageArrayMomentum[0];
  } 
}
