/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_ParticleData.h"


#include "rumd_technical.h"

#include "md_kernel.h"

#include "rumd_utils.h"
#include <cstdio>
#include <cassert>

//////////////////////////////////////////////////////////////////
///  CONSTRUCTION/DESTRUCTION, MEMORY ALLOCATION/DEALLOCATION ////
//////////////////////////////////////////////////////////////////

ParticleData::ParticleData( unsigned int set_particles_per_block, 
			    unsigned int set_threads_per_particle
			    ) {
  
  threads_per_particle = set_threads_per_particle;

  particles_per_block = set_particles_per_block;
  
  
  // these need to be zero before Allocate functions are first called
  allocatedNumParticles = 0;
  //NB allocatedNumBlocksParticles = 0;
}

void ParticleData::SetNumberOfParticles(int set_numParticles) {
  numParticles = set_numParticles;

  printf("\nSetting number of particles and allocating on host and device:\n");
  printf("  Number of particles (n):         %d\n", set_numParticles);
  printf("  Particles per block (pb):  %d\n", particles_per_block);
  printf("  Threads per particle (tp): %d\n", threads_per_particle);
  
  numBlocks = (numParticles+particles_per_block-1)/particles_per_block;    
  printf("  Number of blocks:                %d\n", numBlocks);
  assert( numParticles <= particles_per_block*numBlocks );
  
  numVirtualParticles = numBlocks*particles_per_block;
  printf("  Virtual particles:         %d\n", numVirtualParticles);
  printf("  Unused virtual particles:  %d\n", numVirtualParticles - numParticles);
  assert(numVirtualParticles >= numParticles);
  assert(numVirtualParticles <  numParticles + particles_per_block);
  
  printf("  Threads per block:         %d\n", threads_per_particle*particles_per_block);
  printf("  Number of threads:               %d\n\n", threads_per_particle*particles_per_block*numBlocks);

  // Check parameters;
  assert( threads_per_particle*particles_per_block <= 512 ); // Hardware limit
  assert( particles_per_block <= 32 );                   // (temporary) restriction in nb-list encoding     
  
  // memory that depends on the number of particles
  AllocateParticles(numVirtualParticles);
  // memory that depends on both number of blocks and the number of particles
  //NB AllocateBlocksParticles(numVirtualParticles * numBlocks);

}


void ParticleData::AllocateParticles(int nvp) {
  if( allocatedNumParticles == nvp)
    return;
  FreeParticles();
  
  
  // ===== Allocate host memory =====

  h_r = (float4*) malloc( sizeof(float4) * nvp );
  h_v  = (float4*) malloc( sizeof(float4) * nvp );
  h_vb = (float4*) malloc( sizeof(float4) * nvp );
  
  h_f  = (float4*) malloc( sizeof(float4) * nvp );
  h_w  = (float4*) malloc( sizeof(float4) * nvp );
  h_im = (float4*) malloc( sizeof(float4) * nvp );
  h_Type = (unsigned int*)  malloc( sizeof(unsigned int) * nvp ); 
  h_sts = (float4*) malloc( sizeof(float4) * nvp ); 

  // ===== Allocate device memory =====

  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_r, sizeof(float4)*nvp ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_v, sizeof(float4)*nvp ) );
  //d_vb = &d_v[nvp];   //
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_vb, sizeof(float4)*nvp ) );

  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_f, sizeof(float4)*nvp ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_w, sizeof(float4)*nvp ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_im, sizeof(float4)*nvp ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_sts, sizeof(float4)*nvp ) );

  allocatedNumParticles = nvp;
}

/*NB
void ParticleData::AllocateBlocksParticles(int nb_nvp) {
  if(allocatedNumBlocksParticles == nb_nvp)
    return;
  FreeBlocksParticles();


  // ===== Host memory =====
  h_CNbList = (NBTYPEv*) malloc( sizeof(NBTYPEv) * nb_nvp );

  // ===== Device memory =====
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_CNbList, sizeof(NBTYPEv) * nb_nvp ) );

  // ===== Allocate device memory for nb-list =====

  CUDA_SAFE_CALL( cudaMalloc( (void**) &NbList, sizeof(NBTYPEv) * nb_nvp ) );  
  
  allocatedNumBlocksParticles = nb_nvp;
  
}
*/

void ParticleData::FreeParticles() {
  if(allocatedNumParticles == 0)
    return;
  free(h_r);
  free(h_v);
  free(h_vb);
  
  free(h_f);
  free(h_w);
  free(h_im);
  free(h_Type);
  free(h_sts);

  CUDA_SAFE_CALL(cudaFree(d_r));
  
  CUDA_SAFE_CALL(cudaFree(d_v));
  CUDA_SAFE_CALL(cudaFree(d_vb));
  
  CUDA_SAFE_CALL(cudaFree(d_f));
  CUDA_SAFE_CALL(cudaFree(d_w));
  CUDA_SAFE_CALL(cudaFree(d_im));
  CUDA_SAFE_CALL(cudaFree(d_sts));  
}

/*NB
void ParticleData::FreeBlocksParticles() {
  if(allocatedNumBlocksParticles == 0)
    return;
  
  free(h_CNbList);
  
  CUDA_SAFE_CALL(cudaFree(d_CNbList));

  CUDA_SAFE_CALL(cudaFree(NbList));
  }
*/

ParticleData::~ParticleData() {
  FreeParticles();
  //NB FreeBlocksParticles(); 
}




// ===== HOST TO DEVICE FUNCTIONS  ("Copy-To") =====

void ParticleData::CopyPosToDevice() {
  CUDA_SAFE_CALL( cudaMemcpy( d_r, h_r, sizeof(float4)*numVirtualParticles, cudaMemcpyHostToDevice ) );
}

void ParticleData::CopyVelToDevice() {
  CUDA_SAFE_CALL( cudaMemcpy( d_v, h_v, sizeof(float4)*numVirtualParticles, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_vb, h_vb, sizeof(float4)*numVirtualParticles, cudaMemcpyHostToDevice ) );  
}

void ParticleData::CopyForToDevice() {
   CUDA_SAFE_CALL( cudaMemcpy( d_f, h_f, sizeof(float4)*numVirtualParticles, cudaMemcpyHostToDevice ) );
}


void ParticleData::CopyImagesToDevice() {
  CUDA_SAFE_CALL( cudaMemcpy( d_im, h_im, sizeof(float4)*numVirtualParticles, cudaMemcpyHostToDevice ) );
  
}

/*NB 
void ParticleData::CopyConstraintsToDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( d_CNbList, h_CNbList, sizeof(NBTYPEv) * numVirtualParticles * numBlocks, cudaMemcpyHostToDevice ) );
}
*/

// ===== Wrapper function (bundles all the Copy-To functions) =====

void ParticleData::CopyConfToDevice(){

  CopyPosToDevice();
  CopyVelToDevice();
  CopyForToDevice();
  CopyImagesToDevice();
}


// ===== DEVICE TO HOST FUNCTIONS ("Copy-From") =====

void ParticleData::CopyPosFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_r, d_r, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}

void ParticleData::CopyVelFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_v, d_v,  sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
  CUDA_SAFE_CALL( cudaMemcpy( h_vb, d_vb, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}

void ParticleData::CopyForFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_f, d_f, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}

void ParticleData::CopyVirFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_w, d_w, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}


void ParticleData::CopyImagesFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_im, d_im, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}

void ParticleData::CopyStressFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_sts, d_sts, sizeof(float4)*numVirtualParticles, cudaMemcpyDeviceToHost) );
}

// ===== Wrapper function (bundles all the Copy-From functions) =====

void ParticleData::CopyConfFromDevice(){
  CopyPosFromDevice();
  CopyVelFromDevice();
  CopyForFromDevice();
  CopyVirFromDevice();
  CopyImagesFromDevice();
  CopyStressFromDevice();
}
