/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_sample.h"
#include "rumd_technical.h"

__constant__ __device__ float d_params[MaxNumTypes][MaxNumTypes][NumParam];

#define D_PARAMS
#include "md_kernel.h"

#include <cstdio>
#include <cassert>
#include <iostream>
#include <sstream>
#include <iomanip>

#include "rumd_utils.h"

//////////////////////////////////////////////////////////////////
////////////////// CONSTRUCTION/DESTRUCTION //////////////////////
//////////////////////////////////////////////////////////////////

Sample::Sample( unsigned int set_particles_per_block, unsigned int set_threads_per_particle ) 
  : particleData(set_particles_per_block, set_threads_per_particle),
    trajectoryDir("./TrajectoryFiles"),
    neighborList(0), // The default value has no meaning.
    maxNumIoBlocks(9999){
  
  potential = 0;
  threads_per_particle = set_threads_per_particle;
  particles_per_block = set_particles_per_block;
  
  /////////////////////////////////////////////////////////
  //////////////////// I/O ////////////////////////////////
  /////////////////////////////////////////////////////////
  
  ioformat = 1; // Input-output format (like a version number)
  gz_energyFile = 0; // possibly include in OutputFormat (which would then become more than just a specifier for the format)
  writers_initialized = false;
  newEnergyFile = true; // meaningless until a file has been opened
  writeTrajectoryConfs = true;
  
  outputFormat.writeInterval = 100;
  outputFormat.precision = 6;
  
  trajectoryFormat.precision = 4;
  trajectoryFormat.ioformat = ioformat;
  trajectoryFormat.has_images = true;
  trajectoryFormat.has_velocities = false;
  trajectoryFormat.has_forces = false;
  trajectoryFormat.has_pot_energies = true;
  trajectoryFormat.has_virials = true;
  trajectoryFormat.has_logLin = true;
  
  restartFormat.precision = 9;
  restartFormat.ioformat = ioformat;
  restartFormat.has_images = true;
  restartFormat.has_velocities = true;
  restartFormat.has_forces = true;
  restartFormat.has_pot_energies = true;
  restartFormat.has_virials = true;
  
  // These need to be zero before Allocate functions are first called
  allocatedNumBlocks = 0;
  allocatedNumParticles = 0;
}

Sample::~Sample(){
  FreeAll();
  if(gz_energyFile) gzclose(gz_energyFile);
}

//////////////////////////////////////////////////////////////////
////////////////// MEMORY ALLOCATION /////////////////////////////
//////////////////////////////////////////////////////////////////

void Sample::SetNumberOfParticles(int set_num_part){
  num_part = set_num_part;
  g = 3 * num_part - 3; // Degrees of freedom in the system. 

  particleData.SetNumberOfParticles(set_num_part);
  num_virt_part = particleData.numVirtualParticles;
  num_blocks = particleData.numBlocks;
 
  AllocateParticles(num_virt_part); // Memory that depends on the number of particles
  AllocateBlocks(num_blocks); // Memory that depends on the number of blocks
  AllocateBlocksParticles(num_virt_part * num_blocks); // Memory that depends on both number of blocks and the number of particles

  /////// CUDA execution parameters. ////////
  grid.x = num_blocks; grid.y = 1;
  threads.x = particles_per_block; threads.y = threads_per_particle;
  threads_tp1.x = particles_per_block; threads_tp1.y = 1;
  
  shared_size = particles_per_block * threads_per_particle * sizeof(float4);
  shared_size_tp1 = particles_per_block * sizeof(float);
  
  neighborList.Initialize(&particleData, grid, threads, shared_size, num_blocks, num_virt_part);
}

void Sample::AllocateBlocks(int nb){
  if(allocatedNumBlocks == nb)
    return;
  
  FreeBlocks();
  
  h_Pv                 = (float*) malloc( sizeof(float) * nb );
  h_L                  = (float*) malloc( sizeof(float) * nb );
  h_L[0]               = 0.; // Use a check to see if box length has been set
  h_invL               = (float*) malloc( sizeof(float) * nb );
  h_V                  = (float*) malloc( sizeof(float) * nb );

  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_Pv,                 sizeof(float) * nb ));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_L,                  sizeof(float) * nb ));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_invL,               sizeof(float) * nb ));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_V,                  sizeof(float) * nb ));
    
  allocatedNumBlocks = nb;
}

void Sample::AllocateParticles(int nvp){
  if( allocatedNumParticles == nvp)
    return;
  
  FreeParticles();

  h_misc   = (float4*) malloc( sizeof(float4) * nvp ); 
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_misc,      sizeof(float4) * nvp ));
  
  allocatedNumParticles = nvp;
}

// Empty...?
void Sample::AllocateBlocksParticles(int nb_nvp){
}

//////////////////////////////////////////////////////////////////
////////////////// MEMORY DEALLOCATION ///////////////////////////
//////////////////////////////////////////////////////////////////

void Sample::FreeBlocks(){
  if(allocatedNumBlocks == 0)
    return;
  
  free(h_Pv);
  free(h_L);
  free(h_invL);
  free(h_V);
  
  CUDA_SAFE_CALL(cudaFree(d_Pv));
  CUDA_SAFE_CALL(cudaFree(d_L));  
  CUDA_SAFE_CALL(cudaFree(d_invL));
  CUDA_SAFE_CALL(cudaFree(d_V));
}

void Sample::FreeParticles(){
  if(allocatedNumParticles == 0)
    return;
}

// Empty...?
void Sample::FreeBlocksParticles(){
}

void Sample::FreeAll(){
  FreeBlocks();
  FreeParticles();
  FreeBlocksParticles(); 
}

//////////////////////////////////////////////////////////////////
//////////// HELPER FUNCTIONS FOR INITIALIZATION /////////////////
//////////////////////////////////////////////////////////////////

void Sample::InitializeBoxVariables(float L){
  // Update box and regulator variables for copying.
  assert(L > 0.);
  for(unsigned int i=0; i < num_blocks; i++){
    h_L[i] = L;
    h_V[i] = L * L * L;
    h_invL[i] = 1.f / L;
    h_Pv[i] = h_Pv[0];    
  }
}

void Sample::SetPotential(Potential* set_pot, double set_skin){
  // Set potential and skin
  potential = set_pot;
  skin = set_skin;
      
  // Copy parameters to Device
  printf("Copying interaction parameters to GPU\n\n");
  CUDA_SAFE_CALL(cudaMemcpyToSymbol(d_params, potential->h_params, sizeof(float)*MaxNumTypes*MaxNumTypes*NumParam));
  
  neighborList.SetCutoffs(potential, set_skin);
}

///////////////////////////////////////////////////////////////
////////////  WRAPPERS FOR KERNEL FUNCTIONS ///////////////////
///////////////////////////////////////////////////////////////

void Sample::UpdateNBlist(){ 
  neighborList.UpdateNBlist(d_L, d_invL);
}

void Sample::UpdateNBlistInit(){
  neighborList.UpdateNBlistInit(d_L, d_invL);
}

#include "md_sample_CalcF.cu"

__global__ void set_all_velocities_zero( unsigned int num_part, float4* v ){
  float4 my_v = v[MyGP];
  
  if (MyGP < num_part){
    my_v.x = 0.; my_v.y = 0.; my_v.z = 0.; v[MyGP] = my_v;
  }
}

void Sample::ZeroAllVelocities(){
  float4* d_v = particleData.d_v;
  set_all_velocities_zero<<< grid, threads, shared_size >>>( num_part, d_v );
}

//////////////////////////////////////////////////////////////////
////// FUNCTIONS FOR TRANSFERRING DATA TO THE DEVICE /////////////
//////////////////////////////////////////////////////////////////

void Sample::CopyBoxToDevice() {
  CUDA_SAFE_CALL( cudaMemcpy( d_L, h_L,        sizeof(float) * num_blocks, cudaMemcpyHostToDevice ));
  CUDA_SAFE_CALL( cudaMemcpy( d_invL, h_invL,  sizeof(float) * num_blocks, cudaMemcpyHostToDevice ));
  CUDA_SAFE_CALL( cudaMemcpy( d_V, h_V,        sizeof(float) * num_blocks, cudaMemcpyHostToDevice ));
}

void Sample::CopyConfToDevice(){
  particleData.CopyConfToDevice();
  CUDA_SAFE_CALL( cudaMemcpy( d_Pv, h_Pv,     sizeof(float) * num_blocks, cudaMemcpyHostToDevice ));  
  CopyBoxToDevice();
}

//////////////////////////////////////////////////////////////////
////// FUNCTIONS FOR TRANSFERRING DATA FROM THE DEVICE //////////
//////////////////////////////////////////////////////////////////

void Sample::CopyMiscFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_misc, d_misc, sizeof(float4) * num_virt_part, cudaMemcpyDeviceToHost));
}

void Sample::CopyVolumeFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_V, d_V, sizeof(float), cudaMemcpyDeviceToHost));
}

void Sample::CopyLFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_L, d_L, sizeof(float), cudaMemcpyDeviceToHost));
}

void Sample::CopyinvLFromDevice(){
  CUDA_SAFE_CALL( cudaMemcpy( h_invL, d_invL, sizeof(float), cudaMemcpyDeviceToHost));
}

void Sample::CopyDataFromDevice(){
  particleData.CopyConfFromDevice();
  CopyMiscFromDevice();
  CopyLFromDevice();
  CopyinvLFromDevice();
  CopyVolumeFromDevice();
}

//////////////////////////////////////////////////////////////////
//////////////// VARIOUS HELPER FUNCTIONS ////////////////////////
//////////////////////////////////////////////////////////////////

// Calculate sampled data.
void Sample::CalcEnergies(float4* sumU, float4* sumK, float4* sumW, float4* sumS, float4* sumC, float *stress){
  double U = 0, W_vdw = 0;
  double stress_tmp[6] = {0., 0., 0., 0., 0., 0.};
  double2 config_temp = { 0, 0 };

  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w;
  float4* h_sts = particleData.h_sts;

  // maybe check using counters, maybe only copy. specifically what's needed
  CopyDataFromDevice(); 
  
  for ( unsigned int i = 0; i < num_part; i++ ){ 
    U += h_f[i].w;
    W_vdw += h_w[i].w;
      
    stress_tmp[0] += h_sts[i].x;
    stress_tmp[1] += h_sts[i].y;
    stress_tmp[2] += h_sts[i].z;
    stress_tmp[3] += h_sts[i].w;
    stress_tmp[4] += h_w[i].y;
    stress_tmp[5] += h_w[i].z;
    
    // Nom and denom for configurational temperature.
    config_temp.x += h_f[i].x * h_f[i].x + h_f[i].y * h_f[i].y + h_f[i].z * h_f[i].z;
    config_temp.y += h_misc[i].x;
  }  
  ///////////////// end loop over particles 
  
  // Temperature
  (*sumK).y = ( 2.0 * itg->GetKineticEnergy() ) / (double) g;
  (*sumC).x = config_temp.x;
  (*sumC).y = -config_temp.y;
  (*sumK).w = itg->GetKineticEnergy();     
  (*sumU).w = U * potential->potential_factor / 2.0;
  // Total Van der Waals virial. Factor of 2 for double counting in forces.
  (*sumW).w = ( W_vdw / 2.0 ) / 3.0;
  // Total stress
  for(unsigned int j = 0; j < 6; j++)
    stress[j] = 0.5*stress_tmp[j];
}

void Sample::DumpPotentials() {
  assert (potential != 0);
  
  float4 my_r = {0.0f, 0.0f, 0.0f, 0.0f};
  float4    r = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f}; 
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f}; 
  
  // We don't want to assume member L has been initialized
  float L, invL;
  if(allocatedNumBlocks > 0 && h_L[0] > 0.) {
    L = h_L[0];
    invL = h_invL[0];
  }
  else {
    L = 10.;
    invL = 1.f/L;
  }
 
  FILE *fi; fi = fopen("potentials.dat","w"); 
  
  for (unsigned int i=1; i<999; i++) {
    my_r.x = (L*i)/1000.;
    fprintf(fi, "%f ", my_r.x);
    
    for (unsigned int MyType=0; MyType<MaxNumTypes; MyType++) {
      for (unsigned int Type=0; Type<MaxNumTypes; Type++) {
	my_f.w = 0.0f; 
        potential->fij_cpu(my_r, r, my_r, my_r, &my_f, &my_w, MyType, Type, L, invL, 0);
	fprintf(fi, "%f ", my_f.w*potential->potential_factor);
      }
    }
    fprintf(fi, "\n");
  }   
  fclose(fi);
}

float rnd() {return  rand() / (float) RAND_MAX;}

void Sample::CreateRandomConf(unsigned int num_part, float L, float frac_A, float T) {
  assert(potential != 0);
  SetNumberOfParticles(num_part);

  float4* h_r = particleData.h_r;
  float4* h_im = particleData.h_im;
  float4* h_v = particleData.h_v;
  float4* h_f = particleData.h_f;
  unsigned int* h_Type = particleData.h_Type;

  for(unsigned int tdx = 0; tdx<MaxNumTypes; ++tdx)
    massOfType[tdx] = 1.0;

  
  if(frac_A < 0.0 || frac_A > 1.0)
    fprintf(stderr, "Invalid frac_A: %f\n", frac_A);
  if(frac_A < 1.0)
    this->numTypes = 2;
  else
    this->numTypes = 1;


  // this is required for the memcpy trick with Types
  assert(sizeof(int) == sizeof(float));
  

  // initalize host memory with a random config
  for( unsigned int i = 0; i < num_part; ++i) {
    
    // Set type
    //if (rnd()<0.8) {
    if (i<frac_A*num_part)
      h_Type[i]  = 0;
    else
      h_Type[i]  = 1; 
    // copy (as float) to the device
    memcpy((void*)&(h_r[i].w), (void*)&(h_Type[i]),sizeof(int));
    
    // set image indices (to zero)
    h_im[i].x = 0.0f; h_im[i].y = 0.0f; h_im[i].z = 0.0f;
    // Set velocity (to zero)
    h_v[i].x = 0.0f; h_v[i].y = 0.0f; h_v[i].z = 0.0f;
	// Set mass (to one)
    h_v[i].w = massOfType[h_Type[i]];
    
    // Set forces and energy to zero
    h_f[i].x = 0.0f; h_f[i].y = 0.0f; h_f[i].z = 0.0f; h_f[i].w = 0.0f;
             
    // Place particles at random with 'boltzmann' statistics
    float Ui=1e30;
    int count =0;
    int maxCount = 1000;
    while ( rnd() > exp(-Ui/(T)) && count < maxCount) {
      h_r[i].x = (rnd() - .5)*L; //  -L/2 < x < L/2
      h_r[i].y = (rnd() - .5)*L; 
      h_r[i].z = (rnd() - .5)*L; 
      
      float4 my_f =  {0.0f, 0.0f, 0.0f, 0.0f};
      float4 my_w =  {0.0f, 0.0f, 0.0f, 0.0f};
            
      for (unsigned int j=0; j<i; j++) {
        potential->fij_cpu( h_r[i], h_r[j], h_r[j], h_r[j], &my_f, &my_w, h_Type[i], h_Type[j], L, 1.f/L, 0);
      }
      Ui = my_f.w * potential->potential_factor / 2.0;;
      count += 1;
    }
  }
  
  h_Pv[0] = 0.;

  InitializeBoxVariables(L);
  CopyConfToDevice(); 
}

// Calculate the CM of the molecule. The first particle defines the molecule.
float4 Sample::CalculateCM( float4* pos_array, float4* velo_array, int index ){
  
  float M = 1.f / velo_array[index].w;
  float L = h_L[0];
  
  float4 first = pos_array[index];
  
  float4 CM = { M*first.x, M*first.y, M*first.z, 0 }; 

  int ppm = 1;

  for(int i=index+1; i < (index + ppm); i++){ 
    float mass = 1.f / velo_array[i].w;
    
    // Vector from the first particle to the next particle.
    float3 diff = { pos_array[i].x - first.x, pos_array[i].y - first.y, pos_array[i].z - first.z };
    
    diff.x -= L*rint( diff.x/L );
    diff.y -= L*rint( diff.y/L );
    diff.z -= L*rint( diff.z/L );
    
    // The correct position of the next particle.
    float3 new_pos = { first.x + diff.x, first.y + diff.y, first.z + diff.z };
    
    CM.x += mass * new_pos.x;
    CM.y += mass * new_pos.y;
    CM.z += mass * new_pos.z;
    
    M += mass; 
  }
  
  CM.x /= M;
  CM.y /= M;
  CM.z /= M;
  CM.w = ( 1.0 / M );
  
  return CM;
}

// Isotropic scaling of molecular center of masses. Assumes the same number of particles pr. molecule.
void Sample::IsotropicScaleSystem( float Rscal ){
  CopyLFromDevice();
  particleData.CopyPosFromDevice();
  float4* h_r = particleData.h_r;
  float4* h_v = particleData.h_v;
  float L = h_L[0];

  int ppm = 1;

  for(unsigned int i=0; i < (num_part/ppm); i++){
    int index = i * ppm;

    float4 CM = CalculateCM( h_r, h_v, index );
    
    // Scale coordinates.
    for(int j=index; j < (index + ppm); j++){
      float3 diff = { h_r[j].x - CM.x, h_r[j].y - CM.y, h_r[j].z - CM.z };
      
      diff.x -= L*rint(diff.x/L);
      diff.y -= L*rint(diff.y/L);
      diff.z -= L*rint(diff.z/L);
      
      // The new position.
      float4 new_r = { Rscal * CM.x + diff.x, Rscal * CM.y + diff.y, Rscal * CM.z + diff.z, h_r[j].w };
      
      h_r[j] = new_r;
    }
  }

  InitializeBoxVariables(L * Rscal);
  
  CopyBoxToDevice();
  particleData.CopyPosToDevice();
}

#include "md_sample_io.cu"
