#ifndef RUMD_POTENTIALS_H
#define RUMD_POTENTIALS_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <cassert>
#include <cstdio>
#include <cstring>
#include <cmath>

class Potential {
 public: 
  Potential() {memset(h_params, 0, MaxNumTypes * MaxNumTypes * NumParam * sizeof(float));}
  virtual ~Potential() {};
  double potential_factor;

  float  h_params[MaxNumTypes][MaxNumTypes][NumParam];
  float  h_shell2[MaxNumTypes*MaxNumTypes];
  
  virtual float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) { return 0.f; };
  
  float fij_cpu( float4 my_r, float4 rj, float4 my_cm, float4 qjcm, float4* my_f, float4* my_w,    
		 unsigned int type1, unsigned int type2, float L, float invL, int BS ){
    
    float* param = h_params[type1][type2];
    
    float dr_x = my_r.x - rj.x;
    float dr_y = my_r.y - rj.y;
    float dr_z = my_r.z - rj.z;
    
    dr_x -= L*rintf(dr_x*invL);
    dr_y -= L*rintf(dr_y*invL);
    dr_z -= L*rintf(dr_z*invL);
    
    float dist2 = dr_x * dr_x + dr_y * dr_y + dr_z * dr_z;
    
    // Interaction?
    if( dist2 <= param[0] && dist2 >= 0.000001f ){
      float s = ComputeInteraction_cpu(dist2, param, my_f, my_w);
      
      (*my_f).x += dr_x * s;
      (*my_f).y += dr_y * s;
      (*my_f).z += dr_z * s;
      
      /****** Molecular virial ******/
      if( BS ){
	float3 diff = { my_cm.x - qjcm.x, my_cm.y - qjcm.y, my_cm.z - qjcm.z };
	// Change this for non-cubic volume.
	diff.x -= rintf(diff.x);
	diff.y -= rintf(diff.y);
	diff.z -= rintf(diff.z);
	(*my_w).x += s * ( dr_x * diff.x + dr_y * diff.y + dr_z * diff.z );
      }
    } // END if (dist2< param.z ...
    
    return dist2;
  }
};

class Pot_LJ_12_6 : public Potential {
// Standard 12-6 Lennard-Jones (cut and shifted potential)
// v(r) =  4 Epsilon ( (Sigma/r)^12 -    (Sigma/r)^6 ) - v_cut
// s(r) = 48 Epsilon ( (Sigma/r)^14 - 0.5(Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 

  Pot_LJ_12_6() {
    potential_factor = 1.0 / 12.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 48.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 48.f * Epsilon * ( pow(Rcut,-12.f) - pow(Rcut,-6.f) );   // Potential at Rcut
                               
    printf("Pot_LJ_12_6: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){

    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float s =   param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 0.5f ); // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - param[3];           // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 0.5f );                      // W_ij (virial)
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };       
};


class Pot_gLJ_m_n : public Potential {
// Generalized Lennard-Jones (cut and shifted potential)
// v(r) =   Epsilon/(m -n) ( n*(Sigma/r)^m -  m*(Sigma/r)^n ) - v_cut
// s(r) =  Epsilon ( (Sigma/r)^14 - 0.5(Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 
 float m, n;   // exponents */

  Pot_gLJ_m_n(float set_m, float set_n) {
    m = set_m;
    n = set_n;
    potential_factor = 1.0f / (m*n);
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(6<=NumParam); // Make sure there is allocated space for the needed number of parameters
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = m*n / (m-n) * Epsilon;                                   // Prefactor for forces
    h_params[i][j][3] = n*powf(Rcut,-m) - m*powf(Rcut,-n);                       // "Potential" at Rcut
    h_params[i][j][4] = m*0.5f;                                                  // m/2.0
    h_params[i][j][5] = n*0.5f;                                                  // n/2.0
                              
    printf("Pot_gLJ_m_n: %f %f %f %f %f %f\n", m, n, Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist_m = exp(log(invDist2)*param[4]); 
    float invDist_n = exp(log(invDist2)*param[5]);
    float s =   param[1] * param[2] * (invDist_m - invDist_n) * invDist2; 
    (*my_f).w += param[2] * (2.0f*param[5]*invDist_m - 2.0f*param[4]*invDist_n - param[3]);       // U_ij/potential_factor   
    (*my_w).w += param[2] * (  invDist_m -   invDist_n); 
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };       
};

class Pot_LJ_12_6_NEW : public Potential {
// Standard 12-6 Lennard-Jones (cut and shifted potential) with definition of Sigma from Hoang and Odagasaki
// Sigma HO = 2^(1/6) Sigma LJ
// v(r) =   Epsilon ( (Sigma/r)^12 -  2 *  (Sigma/r)^6 ) - v_cut
// s(r) = 12 Epsilon ( (Sigma/r)^14 - (Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 

  Pot_LJ_12_6_NEW() {
    potential_factor = 1.0 / 12.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma ;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 12.f * Epsilon * ( pow(Rcut,-12.f) - 2.f * pow(Rcut,-6.f) );   // Potential at Rcut
                               
    printf("Pot_LJ_12_6_NEW: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float s =   param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f ); // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 2.0f ) - param[3];           // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 1.0f );                      // W_ij (virial)
    return s;
  }

 __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };
};

class Pot_LJ_12_6_Gauss : public Potential {
// 12-6 Lennard-Jones plus Gaussian potential cut and shifted.
// v(r) =  Epsilon ( (Sigma/r)^12 -  2(Sigma/r)^6 - Epsilon_0 (exp[-(r - r_0)^2 / (2 Sigma_0^2)] ) ) - v_cut
// v(r) is a double well with second well located at r_0 with depth Epsilon_0.

// s(r) = 48 Epsilon ( (Sigma/r)^14 - 0.5(Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 

  Pot_LJ_12_6_Gauss() {
    potential_factor = 1.0 / 12.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut, float Sigma0, float Epsilon0, float r0) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(7<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                              // (Rcut_ij)^2  (Rcut := rcut/Sigma)
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                 // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                           // Prefactor for forces 
    h_params[i][j][3] = 12.f * Epsilon * ( pow (Rcut,-12.f) - 2.f *  pow (Rcut,-6.f) 
					 - (Epsilon0 * exp (- pow ((Rcut * Sigma - r0),2.f) / ( 2.f * pow (Sigma0,2.f))))); // Potential at Rcut
    h_params[i][j][4] = 2.0f * Sigma0 * Sigma0;                                   // Sigma0
    h_params[i][j][5] = r0;                                                       // r0
    h_params[i][j][6] = Epsilon0;                                                 // Epsilon0
                               
    printf("LJ_12_6_Gauss: %f %f %f %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3], Sigma0, r0, Epsilon0);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float s =    param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f )
                 - (param[6] * param[2] * ( sqrt (dist2) - param[5] ) / ( 6.0f * sqrt (dist2) *  param[4])) 
                 * exp (- pow ((sqrt (dist2) - param[5]),2.f) / param[4]);                  // F_ij = s * r_ij
    (*my_f).w += param[2] * (invDist6 * ( invDist6 - 2.0f )
			      - param[6] * exp (- pow ((sqrt (dist2) - param[5]),2.0f) / param[4])) - param[3];   // U_ij
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 1.0f )
			      - (param[6] * param[2] * ( dist2 - sqrt (dist2) * param[5] ) / ( 6.0f *  param[4])) 
                              * exp (- pow ((sqrt (dist2) - param[5]),2.f) / param[4]) ;         // W_ij (virial)
    return s;
  }

 __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };
        

};

class Pot_Gauss : public Potential {
// Gaussian core potential
// v(r) = Epsilon * exp[-(r/Sigma)^2]
// s(r) = ( 2 * Epsilon / Sigma^2 ) * exp[-(r/Sigma)^2] = -r^-1*dv/dr (F_ij = s * r_ij)
 public: 

  Pot_Gauss() {
    potential_factor = 1.0 / 1.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma ;                             // (Rcut_ij)^2       Rcut := rcut/Sigma
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = Epsilon * exp (-( Rcut * Rcut ));   // Potential at Rcut
                               
    printf("Pot_Gauss: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float Dist2 = dist2 * param[1];                                // (r/Sigma)^2 
    float s = 2.f * param[1] * param[2] * exp(-Dist2);             // F_ij = s * r_ij
    (*my_f).w += param[2] * exp(- Dist2) - param[3];               // U_ij 
    (*my_w).w += 2.f * Dist2 * param[1] * param[2] * exp(-Dist2);  // W_ij (virial) = s * r^2
    return s;
  }

 __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };
        

};


class Pot_Buckingham3 : public Potential {
// Buckingham potential
// v(r) = Epsilon{(6/(Alpha - 6))*exp[Alpha(1-r/rm)] - (Alpha/(Alpha -6))*(rm/r)**6}
// s(r) = -r^-1*dv/dr = ((6*Alpha*Epsilon)/(Alpha - 6)){r^(-1)*exp[Alpha(1-r/rm)] - rm^(-6)*r^(-8)}
// w(r) = s*r^2 = ((6*Alpha*Epsilon)/(Alpha - 6))*{r*exp[Alpha(1-r/rm)] - (rm/r)^6}
 public: 

  Pot_Buckingham3() {
    potential_factor = 1.0;
  }

  void SetParams(unsigned int i, unsigned int j, float rm, float Alpha, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(6<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut;                                             // (Rcut_ij)^2
    h_params[i][j][1] = Epsilon;                                                 // Prefactor
    h_params[i][j][2] = Alpha / (Alpha - 6.f);                                   // Prefactor 
    h_params[i][j][3] = Alpha;                                                   // Prefactor
    h_params[i][j][4] = rm;                                                      // Prefactor 
    h_params[i][j][5] = Epsilon * ( (6.f/(Alpha - 6.f)) * exp (Alpha * (1.f - Rcut/rm) ) - (Alpha/(Alpha -6.f))* pow ((rm/Rcut),6.f) ); // Potential at Rcut
                               
    printf("Pot_Buckingham3: %f %f %f %f %f \n", rm, Alpha, Epsilon, Rcut, h_params[i][j][4]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2 = 1.0f / ( dist2 );                                                      // (1/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                                        // (1/r)^6
    float s =  6.f * param[1] * param[2] * ( (sqrt (invDist2) / param[4]) * exp ( param[3] * (1.f - 1.f /(param[4]*sqrt(invDist2))))
					     - pow ( param[4] , 6.f) * invDist2 * invDist6);    // F_ij = s * r_ij
    (*my_f).w +=  param[1] * ((6.f*(param[2]/param[3])) *exp (param[3] * (1.f - 1.f /(sqrt (invDist2) * param[4]) ) ) 
			      - param[2] * pow (param[4],6.f) * invDist6) - param[5]; // U_ij
    (*my_w).w +=  6.f * param[1] * param[2]*((1.f/(param[4]*sqrt(invDist2)))*exp ( param[3] * (1.f - 1.f /(param[4]*sqrt(invDist2))))
					     -  pow (param[4],6.f) * invDist6);             // W_ij (virial) = s * r^2
    return s;
  }

 __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }


  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };

};


class Pot_IPL_12 : public Potential {
// IPL potential, n=12. (cut and shifted potential)
// v(r) =    Epsilon ( (Sigma/r)^12 ) - v_cut
// s(r) = 12 Epsilon ( (Sigma/r)^14 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 

  Pot_IPL_12() {
    potential_factor = 1.0 / 12.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 12.f * Epsilon * ( pow(Rcut,-12.f) );                    // Potential at Rcut
                               
    printf("Pot_IPL_12: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2  = 1.0f / ( dist2 * param[1] );                               // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;                            // (sigma/r)^6
    float invDist12 = invDist6 * invDist6;                                       // (sigma/r)^12    
          
    float s =   param[1] * param[2] * invDist12 * invDist2;                      // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist12 - param[3];                                // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist12;                                           // W_ij (virial)
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };
};

class Pot_IPL_18 : public Potential {
// IPL potential, n=18. (cut and shifted potential)
// v(r) =    Epsilon ( (Sigma/r)^18 ) - v_cut
// s(r) = 18 Epsilon ( (Sigma/r)^20 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 

  Pot_IPL_18() {
    potential_factor = 1.0 / 18.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 18.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 18.f * Epsilon * ( pow(Rcut,-18.f) );                    // Potential at Rcut
                               
    printf("Pot_IPL_18: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2  = 1.0f / ( dist2 * param[1] );                               // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;                            // (sigma/r)^6
    float invDist18 = invDist6 * invDist6 * invDist6;                            // (sigma/r)^18    
          
    float s =   param[1] * param[2] * invDist18 * invDist2;                      // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist18 - param[3];                                // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist18;                                           // W_ij (virial)
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w);
  };
};





class Pot_IPL_n : public Potential {
// IPL potential, abritrary exponent n. (cut and shifted potential)
// v(r) =   Epsilon ( (Sigma/r)^n ) - v_cut
// s(r) = n Epsilon ( (Sigma/r)^(n+2) )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij)

 public: 
  float n;   // exponent

  Pot_IPL_n() {
    potential_factor = 1.0;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut, float n) {
    assert(i<MaxNumTypes); assert(j<MaxNumTypes);
    assert(4<=NumParam); // Make sure there is allocated space for the needed number of parameters      
          
    // !!! NOTE: (Rcut_ij)^2 in absolute units should ALLWAYS ALLWAYS ALLWAYS be the 0'th parameter !!!
    // Stored potential energies are in units of 'potential_factor'

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = Epsilon;                                                 // Prefactor for forces
    h_params[i][j][3] = Epsilon * ( pow(Rcut,-n) );                              // Potential at Rcut
    h_params[i][j][4] = n;                                                       // n
                               
    printf("Pot_IPL_n=%f: %f %f %f %f\n", n, Sigma, Epsilon, Rcut, h_params[i][j][3]);
  }                                    

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {

    float invDist2   = 1.0f / ( dist2 * param[1] );                              // (sigma/r)^2 
    //float invDist_n  = powf(invDist2, param[4]*0.5);                           // (sigma/r)^n
    float invDist_n  = exp(log(invDist2)*param[4]*0.5);                          // (sigma/r)^n (faster)
                    
    float s =    param[1] * param[2] * param[4] * invDist_n * invDist2;          // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist_n - param[3];                                // U_ij/potential_factor 
    (*my_w).w += param[2] * param[4] * invDist_n;                                // W_ij (virial)
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) {
    return ComputeInteraction(dist2, param, my_f, my_w); 
  };
};

/**********************************************************/
/***************** Smoothed potentials ********************/
/**********************************************************/

// 12-6 Lennard Jones with Gromacs style smoothing.
class Pot_LJ_12_6_Gromacs : public Potential{
  
 public: 
  
  Pot_LJ_12_6_Gromacs() {
    potential_factor = 1.0 / 12.0;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    assert(i<MaxNumTypes); assert(j<MaxNumTypes); assert(9<=NumParam); 
    
    // Hard-coded.
    float Rsmooth = 2.45;
    float diff = Sigma * ( Rcut - Rsmooth );
    
    float Force = ( 48.f * Epsilon * ( pow(Rcut,-13.f) - 0.5 * pow(Rcut,-7.f) ) ) / Sigma;   
    float Deriv = ( - 624.f * Epsilon * pow(Rcut,-14.f) + 168 * Epsilon * pow(Rcut,-8.f) ) / ( Sigma * Sigma ); 
    
    // Gromacs constants.
    float A = ( - 3.f * Force + Deriv * diff ) / powf(diff,2.f);
    float B = ( - Deriv * diff + 2.f * Force ) / powf(diff,3.f);
    float C = ( A / 3.f ) * powf(diff,3.f) + ( B / 4.f ) * powf(diff,4.f);
    
    // GPU parameters.
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                           
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                               
    h_params[i][j][2] = 48.f * Epsilon;                                        
    h_params[i][j][3] = 48.f * Epsilon * ( pow(Rcut,-12.f) - pow(Rcut,-6.f) ); 
    h_params[i][j][4] = Rsmooth * Rsmooth * Sigma * Sigma;
    h_params[i][j][5] = A;
    h_params[i][j][6] = B;
    h_params[i][j][7] = 12.f * C;
    h_params[i][j][8] = Rsmooth * Sigma;
    printf("Pot_LJ_12_6_Gromacs: %f %f %f %f - A = %f; B = %f; C = %f;\n", Sigma, Epsilon, Rcut, Rsmooth, A, B, C);
  }                                    
  
  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    
    float invDist2 = 1.0f / ( dist2 * param[1] );                               
    float invDist6 = invDist2 * invDist2 * invDist2;                            
    
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 0.5f ); 
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - param[3] + param[7];  
        
    // r >= Rsmooth.
    if( dist2 >= param[4] ){
      float r = sqrtf(dist2);
      float diff = r - param[8]; 
      float diff_sq = diff * diff;
    
      s += ( diff_sq * ( param[5] + param[6] * diff ) / r );
      (*my_f).w += diff_sq * ( - 4.f * diff * param[5] - 3.f * diff_sq * param[6] ); 
    }
    
    return s;
  }
  
  // Equivalent to above, but adds extra computation for config temp.
  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){
    
    float invDist2 = 1.0f / ( dist2 * param[1] );                               
    float invDist6 = invDist2 * invDist2 * invDist2;                           
    float common =  param[1] * param[2] * invDist6 * invDist2;
    float s = common * ( invDist6 - 0.5f ); 
    
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - param[3] + param[7];  
    (*my_misc).x += 3.f * s + common * ( -14.f * invDist6 + 4.f );

    // r >= Rsmooth.
    if( dist2 >= param[4] ){
      float r = sqrtf(dist2); 
      float diff = r - param[8]; 
      float diff_sq = diff * diff;
      
      float force = ( diff_sq * ( param[5] + param[6] * diff ) ) / r; 
      s += force;
      
      (*my_f).w += diff_sq * ( - 4.f * diff * param[5] - 3.f * diff_sq * param[6] ); 
      (*my_misc).x += 2.f * force + diff * ( 2.f * param[5] + 3.f * diff * param[6] );
    }
    
    return s;
  }
  
  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){
    return ComputeInteraction(dist2, param, my_f, my_w);
  };   
};

#endif // RUMD_POTENTIALS_H
